/* 
 * Laboratorio di Programmazione - A.A. 2004-05
 * Canali A-D (Nicola Galesi), P-Z (Stefano Guerini)
 * 
 * parser.c
 */

#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include "nodes.h"
#include "parser.h"

#define next_tk(fp, ptk) { if (! next_token(fp, ptk)) return 0; }

/*
 * stampa token
 * PRE: *ptk e' un token
 * POST: *ptk e' immutato
 */
void print_tk(token *ptk) {
    assert(ptk != NULL);
    switch (ptk->kind) {
    case special:
	printf("special: %c\n", ptk->u.ch);
	break;
    case word:
	printf("word: %s\n", ptk->u.str);
	break;
    default: /* valore non valido */
	assert(0);
	break;
    }
}

/*
 * stampa word
 * PRE: *piw e' una struttura con le info della parola
 * POST: *piw e' immutato
 */
void print_word(struct info_word *piw) {
    assert(piw != NULL);
    printf("[%d, %c, %c%c, %c, %d, \"%s\"] \"%s\"\n",
	   piw->word_pos,
	   (piw->title ? 't' : 'b'),
	   (piw->bold ? 'b' : '-'),
	   (piw->italic ? 'i' : '-'),
	   (piw->word_size == small ? 's' :
	    (piw->word_size == large  ? 'l' : 'n')),
	   piw->link_pos,
	   (piw->file_link == NULL ? "NULL" : piw->file_link),
	   piw->parola);
}

/*
 * legge il prossimo carattere diverso da spazio
 */
int next_nb(FILE *fp) {
    int c;
    for (c = fgetc(fp); isspace(c); c = fgetc(fp));
    return(c);
}

/*
 * legge il prossimo token
 * PRE: *ptk e' un token esistente
 * POST: *ptk contiene il token letto
 */
int next_token(FILE* fp, token *ptk) {
    int c = next_nb(fp); /* primo carattere del token */
    int i;

    assert(ptk != NULL);
    if (c == EOF)
	return 0;
    if (! isalnum(c)) {
	ptk->kind = special;
	ptk->u.ch = c;
    } else {
	ptk->kind = word;
	for (i = 0; isalnum(c) && i < MAX_TOKEN_LENGTH; c = fgetc(fp), i++) 
	    ptk->u.str[i] = c; /* appendi il carattere */
	ptk->u.str[i] = '\0';  /* chiudi la stringa */
	c = ungetc(c, fp);  /* restituisci a stdin il carattere
				* che segue la parola letta */
    }
    return -1;
}

/*
 * legge un token racchiuso tra doppi apici
 * il primo doppio apice e' stato letto
 * tutto cio' che occorre in input sino al prossimo doppio apice
 * forma il token, il token restituito ha kind = word
 * PRE: *ptk e' un token esistente
 * POST: *ptk contiene il token letto
 *       ritorna 0 se si incotra EOF prima dei doppi apici o
 *       la parola e' troppo lunga , -1 altrimenti
 */
int next_quoted_word(FILE *fp, token *ptk) {
    int c;
    int i;

    assert(ptk != NULL);
    ptk->kind = word;
    for (i = 0, c = fgetc(fp);
	 i < MAX_TOKEN_LENGTH && c != '"' && c != EOF;
	 i++, c = fgetc(fp))
	ptk->u.str[i] = c; /* appendi il carattere */
    ptk->u.str[i] = '\0';  /* chiudi la stringa */
    if (c != '"')
	return 0;
    else
	return -1;
}


/*
 * legge il prossimo tag
 * PRE: *pit un tag esistente
 *      il < iniziale del tag e' gia' stato letto
 * POST: *pit contiene il tag letto
 */
int next_tag(FILE *fp, struct info_tag *pit) {
    token tk;

    next_tk(fp, &tk);
    if (tk.kind == special && tk.u.ch == '/') { /* tag finale */
	pit->k = close;
	next_tk(fp, &tk);
    } else /* tag iniziale */
	pit->k = open;
    if (tk.kind == word) {
	if (strcasecmp(tk.u.str, "a") == 0) { /* link */
	    if (pit->k == close)
		pit->str[0] = '\0';
	    else {
		pit->t = a_tag;
		next_tk(fp, &tk);
		/* cerca href e ignora il resto */
		while ((tk.kind != special || tk.u.ch != '>') &&
		       (tk.kind != word || 
			strcasecmp(tk.u.str, "href") != 0)) {
		    next_tk(fp, &tk);
		}
		if (tk.kind == word && strcasecmp(tk.u.str, "href") == 0) {
		    /* href trovato */
		    next_tk(fp, &tk); /* leggi = */
		    if (tk.kind == special && tk.u.ch == '=') {
			next_tk(fp, &tk); /* leggi il link */
			if (tk.kind == special && tk.u.ch == '"')
			    /* link racchiuso tra apici */
			    if (! next_quoted_word(fp, &tk))
				/* eof o errore prima della 
				 * chiusura dei doppi apici
				 */ 
				return 0;
		    }
		    if (tk.kind == word) /* link valido, memorizzalo */
			strcpy(pit->str, tk.u.str);
		    else /* link non valido, azzera il link */
			pit->str[0] = '\0';
		}
	    }
	} else if (strcasecmp(tk.u.str, "title") == 0) {
	    pit->t = title_tag;
	} else if (strcasecmp(tk.u.str, "body") == 0) {
	    pit->t = body_tag;
	} else if (strcasecmp(tk.u.str, "small") == 0) {
	    pit->t = small_tag;
	} else if (strcasecmp(tk.u.str, "large") == 0) {
	    pit->t = large_tag;
	} else if (strcasecmp(tk.u.str, "b") == 0) {
	    pit->t = b_tag;
	} else if (strcasecmp(tk.u.str, "i") == 0) {
	    pit->t = i_tag;
	} else 
	    pit->t = unknown_tag;
    } else
	pit->t = unknown_tag;
    /* se ancora non e' stato letto >, cerca il prossimo > 
     * ignorando i token letti prima di trovarlo */
    while (tk.kind != special || tk.u.ch != '>')
	next_tk(fp, &tk);
    return -1;
}


/*
 * legge la prossima parola
 * PRE: *piw una struct info word valida
 * POST: *piw contiene la parola letta
 *       le stringhe assegnate a parola e link sono buffer statici
 *       interni che verranno sovrascritti dalle successive chiamate
 *       della funzione
 */
int next_wd(FILE *fp, struct info_word *piw) {
    static char link[256];
    static char parola[256];
    token tk;
    struct info_tag tag;

    assert(piw);
    do {
	next_tk(fp, &tk);
    } while (tk.kind != word && (tk.kind != special || tk.u.ch != '<'));
    while (tk.kind == special && tk.u.ch == '<') {
	if (! next_tag(fp, &tag))
	   return 0;
	switch (tag.t) {
	case title_tag:
	    piw->title = (tag.k == open);
	    break;
	case body_tag:
	    piw->title = 0;
	    break;
	case a_tag:
	    if (tag.k == open) {
		strcpy(link, tag.str);
		piw->file_link = link;
		piw->link_pos++;
	    } else
		piw->file_link = NULL;
	    break;
	case small_tag:
	    if (tag.k == open)
		piw->word_size = small;
	    else
		piw->word_size = normal;
	    break;
	case large_tag:
	    if (tag.k == open)
		piw->word_size = large;
	    else
		piw->word_size = normal;
	    break;
	case b_tag:
	    piw->bold = (tag.k == open);
	    break;
	case i_tag:
	    piw->italic = (tag.k == open);
	    break;
	default:
	    break;
	}
	do {
	    next_tk(fp, &tk);
	} while (tk.kind != word && (tk.kind != special || tk.u.ch != '<'));
    }
    strcpy(parola, tk.u.str);
    piw->parola = parola;
    piw->word_pos++;
    return -1;
}

/*
 * variabile statica per la lettura di word
 * usata da next_word
 */
static struct info_word iw;

/* valori di default per iw */
static struct info_word iw_defaults = {NULL, 0, 0, 0, 0, normal, NULL, 0};

/* 
 * apre un ipertesto per leggere le parole mediante next_word
 * fname e' il nome del file da aprire
 * ritorna il descrittore del file pronto per la lettura se tutto OK,
 * o NULL in caso di errore
 * POST: azzera tutti gli attributi per le successive chiamate di next_word
 */
FILE *open_ipertesto(char *fname) {
    FILE *fp = fopen(fname, "r");
    iw = iw_defaults;
    return fp;
}

/*
 * chiude un ipertesto fp
 * POST: azzera tutti gli attributi per le successive chiamate di next_word
 */
void close_ipertesto(FILE *fp) {
    iw = iw_defaults;
    fclose(fp);
}

/*
 * legge la prossima parola del file fp
 * PRE: fp e' un descrittore di file ottenuto da una open_ipertesto
 * POST: ritorna una nuova struttura conetenente tutte le informazioni
 *       della parola letta. Tutta la memoria utilizzata dalla nuova
 *       struttura, incluse le stringhe della parola e dell'eventuale
 *       link, e' nuova ed e' stata allocata dalla chiamata alla funzione.
 *       In caso di EOF ritorna NULL
 */
struct info_word *next_word(FILE *fp) {
    struct info_word *piw;

    if (! next_wd(fp, &iw))
	return NULL; /* errore di lettura o eof */
    piw = malloc(sizeof(struct info_word));
    assert(piw);
    *piw = iw; /* copia la struttura della parola */
    /* copia le stringhe della parola e dell'eventuale link */
    if (iw.parola != NULL) {
	piw->parola = malloc(strlen(iw.parola)+1);
	assert(piw->parola);
	strcpy(piw->parola, iw.parola);
    }
    if (iw.file_link != NULL) {
	piw->file_link = malloc(strlen(iw.file_link)+1);
	assert(piw->file_link);
	strcpy(piw->file_link, iw.file_link);
    }
    return(piw);
}
