/* 
 * Laboratorio di Programmazione - A.A. 2004-05
 * Canali A-D (Nicola Galesi), P-Z (Stefano Guerini)
 * 
 * table.h
 */

#if !defined __TABLE_H

#define __TABLE_H 1

#include "hashtab.h"

/*
 * Elemento di una tavola.
 *
 * Il campo key e' la chiave del'elemento, una stringa.
 * Il campo info e' un puntatore alla struttura contenente i dati
 * associati alla chiave. Per poter implementare la tavola indipendentemente
 * dal tipo dei dati associati alla chiave, questo campo e' di tipo void *.
 */
struct tabelem {
    char *key; /* chiave dell'elemento */
    void *info; /* informazioni associate alla chiave */
};

/* 
 * Lista di elementi di una tavola. 
 */
struct telist {
    struct tabelem *el; /* elemento della tavola */
    struct telist *next; /* coda della lista */
};

/*
 * Tavola di elementi.
 *
 * La tavola e' formata da
 * - una tavola hash nella quale gli elementi vengono inseriti e ricercati
 *   per mezzo della chiave
 * - il puntatore alla lista di tutti igli elementi nella tavola
 * NB La lista degli elementi serve a scorrere in modo sequenziale
 *    gli elementi presenti nella tavola.
 */
struct table {
    HASHTABLE htable; /* tavola hash */
    struct telist *ls; /* lista di tutti i nodi */
};

/* 
 * Verifica se una tavola e' piena.
 *
 * Ritorna true (non 0) se la tavola e' piena, 0 altrimenti.
 */
int table_full(struct table *t);

/* 
 * Inizializza una tavola.
 */
void table_init(struct table *t);

/* 
 * Inserisce in una tavola un nuovo elemento.
 *
 * struct table *t: e' la tavola in cui inserire l'elemento
 * char *key: la chiave dell'elemento da inserire
 * Ritorna il puntatore all'elemento inserito o NULL in caso di errore 
 * (eg, tavola piena) o elemento gia' presente. Il campo info del nuovo
 * elemento inserito viene inizializzato a NULL.
 */
struct tabelem *table_ins(struct table *t, char *key);

/* 
 * Cerca in una tavola un elemento.
 *
 * struct table *t: e' la tavola in cui ricercare l'elemento
 * char *key: la chiave dell'elemento da cercare
 * Ritorna il puntatore all'elemento cercato, se questo e' presente nella
 * tavola, o NULL, nel caso in cui l'elemento non e' presente.
 */
struct tabelem *table_search(struct table *t, char *key);

#endif
